;;******************************************************************************
;;                         memory.inc      memory.inc
;;******************************************************************************
;;
;;  Copyright (C) 1990 Vance Morrison
;;
;;
;; Permission to view, compile, and modify for LOCAL (intra-organization) 
;; USE ONLY is hereby granted, provided that this copyright and permission 
;; notice appear on all copies.  Any other use by permission only.
;;
;; Vance Morrison makes no representations about the suitability 
;; of this software for any purpose.  It is provided "as is" without expressed 
;; or implied warranty.  See the copywrite notice file for complete details.
;;
;;******************************************************************************
;; 
;; unfortuately, using the full 16bit bus of an AT with the shared memory
;; of networking cards is not transparent (at least not with the WD8013EBT).
;; because of this, we need to toggle 16bit mode when it is necessary
;; (it is also necessary to do this with interupts disabled, which also
;; requires special processing).  memory.inc provides functions that allow
;; this toggling to be done automatically.  
;;
;;******************************************************************************
;;
;; functions provided by this module are
;;
;;   MEM_DECLARE name
;;   MEM_DEFINE name
;;   MEM_COPY_in_CX_SI_DI_ES_out_SI_DI_const_AX_BX_DX_BP_ES  
;;   MEM_REGISTER_REGION_in_DI_ES enable_code, disable_code
;;
;;******************************************************************************

mem_entry STRUC
    mem_on_routine  DW ?
    mem_off_routine DW ?
mem_entry ENDS

;;******************************************************************************
;;   MEM_DECLARE  declares your intention to use the services of this module
;; 'name' is only used to insure that 'MEM_DEFINE' is only called once
    
MEM_DECLARE MACRO name
    .errb <tag>
    .DATA
    mem_declared = name
    global mem_16bit_table:byte 
    global mem_save_AX:word 
    global mem_save_BX:word 
    global mem_save_DX:word 
    global mem_mem_entry:near 
   .CODE
ENDM


;;******************************************************************************
;;   MEM_DEFINE
;;      sets asside memory for the MEM functions and initializes it.
;;      routine is a no-op if 'name' was not declared
;;
MEM_DEFINE MACRO name
    local around, dummy_entry

ifdef mem_declared
if name eq mem_declared             
    .DATA
    mem_defined = 1
    mem_16bit_table db 512 dup (?)  ;; table holds enable/disable routines
    mem_save_AX dw ?                    ;; space for temps
    mem_save_BX dw ?                    ;; space for temps
    mem_save_DX dw ?                    ;; space for temps

    .CODE
   jmp around
       mem_mem_entry:
      ret
   around:
       mov AX, DS           ;; initialize the table 
       mov ES, AX
       mov AX, offset mem_mem_entry
       mov DI, offset mem_16bit_table
       mov CX, 0A0H         ;; first 80 entries are main memory
       rep stosw
       xor AX, AX           ;; the remaining 48 are unknown
       mov CX, 060H         
       rep stosw
endif
endif
ENDM


;;******************************************************************************
;; MEM_REGISTER_REGION registers the 8K block of memory starting at DI:ES
;; as being potentially 16bit memory.  The code 'enable_code' and 'disable_code'
;; are subroutines (ends with ret), that enable and disable the 16 bit option.
;; Once the memory is registered, MEM_COPY will use 16bit mode if possible.
;; Note that memory below 640K (A0000) is assumed to be main memory and
;; always able to do 16 bit transfers. 
;;
MEM_REGISTER_REGION_in_DI_ES MACRO enable_code, disable_code
    .errb <enable_code>
    .errb <disable_code>

    mov BX, DI
    shr BX, 1
    shr BX, 1
    shr BX, 1
    shr BX, 1
    mov AX, ES 
    add BX, AX
    shr BX, 1
    mov BL, BH
    xor BH, BH
    shl BX, 1               ;; multiply by 4 (the size of the entries)
    shl BX, 1
    add BX, offset mem_16bit_table
    mov [BX+mem_on_routine], offset enable_code
    mov [BX+mem_off_routine], offset disable_code
ENDM


;;******************************************************************************
;; MEM_COPY_in_CX_SI_DI_ES_out_SI_DI does the copy operation.  It requires
;; that SI and DI are EVEN.   This function is esentially a 'rep movsw'
;; with added functionality that enables 16bit bus transfers if possible.
;; Note that MEM_COPY assumes that if the first address in the block is
;; capable of 16 bit transfer, the whole block can be transfered in 16bit mode.
;; This routine is NOT meant to be called from interupt level
;;

MEM_COPY_in_CX_SI_DI_ES_out_SI_DI_const_AX_BX_DX_BP_ES MACRO 
    local done, not_16

        ;; are there any 16 bit cards out there?
    ifdef mem_declared
        mov mem_save_AX, AX
        mov mem_save_BX, BX
        mov mem_save_DX, DX

        mov BX, SI
        shr BX, 1
        shr BX, 1
        shr BX, 1
        shr BX, 1
        mov AX, DS 
        add BX, AX
        shr BX, 1
        mov BL, BH
        xor BH, BH
        shl BX, 1               ;; multiply by 4 (the size of the entries)
        shl BX, 1
        add BX, offset mem_16bit_table
        mov DX, BX
        cmp word ptr [BX], 0             ;; is the source 16 bit capable?
        jz not_16

        mov BX, DI
        shr BX, 1
        shr BX, 1
        shr BX, 1
        shr BX, 1
        mov AX, ES 
        add BX, AX
        shr BX, 1
        mov BL, BH
        xor BH, BH
        shl BX, 1               ;; multiply by 4 (the size of the entries)
        shl BX, 1
        add BX, offset mem_16bit_table
        cmp word ptr [BX], 0             ;; is the dest 16 bit capable?
        jz not_16
                                ;; then do a 16 bit transfer
        cli
        call [BX+mem_on_routine]
        xchg DX, BX
        call [BX+mem_on_routine]
        inc CX
        shr CX, 1
        rep movsw
        call [BX+mem_off_routine]
        xchg DX, BX
        call [BX+mem_off_routine]
        sti
        jmp done

        not_16:
            inc CX
            shr CX, 1
            rep movsw           ;; if not, just copy    

        done:
        mov AX, mem_save_AX
        mov BX, mem_save_BX
        mov DX, mem_save_DX
    else
    inc CX
    shr CX, 1
    rep movsw           ;; if not, just copy    
    endif
ENDM

